/******************************************************************************
Asservissement de verticalité et de suivi de ligne d'un gyropode.
Ce programme est destiné à fonctionner avec Geeros Wifi ou Geeros Lynx, 
disponibles à l'adresse suivante:
http://boutique.3sigma.fr/21-geeros-wifi.html
http://boutique.3sigma.fr/22-geeros-lynx.html

ATTENTION: ce programme fonctionne uniquement avec les robots livrés
avant le 1er juillet 2013

Auteur: 3Sigma
Version 2.1 - 13/11/2013
*******************************************************************************/


// Inclusion d'une bibliothèque permettant l'exécution à cadence fixe
// d'une partie du programme. Télécharger à l'adresse http://www.3sigma.fr/telechargements/FlexiTimer2.zip
// et décompresser dans le sous-répertoire « libraries » de votre installation Arduino
// Pour plus de détails, voir les pages (en anglais): 
// http://www.arduino.cc/playground/Main/FlexiTimer2
// https://github.com/wimleers/flexitimer2
#include <FlexiTimer2.h>

// Inclusion d'une bibliothèque permettant de lire et d'écrire plus rapidement sur les entrées-sorties digitales.
// Télécharger à l'adresse http://www.3sigma.fr/telechargements/digitalWriteFast.zip
// et décompresser dans le sous-répertoire « libraries » de votre installation Arduino
#include <digitalWriteFast.h> 

// Librairie Arduino requise pour communiquer avec des composants sur interface I2C
#include "Wire.h"

// Inclusion d'une bibliothèque permettant de communiquer avec le capteur accélérométrique
// et gyroscopique 3 axes sur interface I2C
// Télécharger à l'adresse http://www.3sigma.fr/telechargements/I2Cdev.zip
// et décompresser dans le sous-répertoire « libraries » de votre installation Arduino
#include "I2Cdev.h"

// Inclusion d'une bibliothèque intègrant de nombreuses fonctions permettant d'accéder aux
// informations mesurées par le MPU6050, qui est le le capteur accélérométrique et
// gyroscopique 3 axes intégré à Geeros.
// Télécharger à l'adresse http://www.3sigma.fr/telechargements/MPU6050.zip
// et décompresser dans le sous-répertoire « libraries » de votre installation Arduino
#include "MPU6050.h"

// Inclusion d'une bibliothèque permettant une commande synchronisée avec l'"interruption timer"
// de l'asservissement (voir plus loin) du servomoteur supportant la caméra
// Télécharger à l'adresse http://www.3sigma.fr/telechargements/SoftwareServo.zip
// et décompresser dans le sous-répertoire « libraries » de votre installation Arduino
#include <SoftwareServo.h> 

// Définitions et déclarations pour le codeur incrémental du moteur droit
#define codeurDroitInterruptionA 0
#define codeurDroitPinA 2
#define codeurDroitPinB 8
#define Nmoy 10
volatile long ticksCodeurDroit = 0;
static int indiceTicksCodeurDroit = 0;
static int ticksCodeurDroitTab[Nmoy];
static int codeurDroitDeltaPos;

// Définitions et déclarations pour le codeur incrémental du moteur gauche
#define codeurGaucheInterruptionA 1
#define codeurGauchePinA 3
#define codeurGauchePinB 9
volatile long ticksCodeurGauche = 0;
static int indiceTicksCodeurGauche = 0;
static int ticksCodeurGaucheTab[Nmoy];
static int codeurGaucheDeltaPos;

// Définitions et déclarations pour le moteur à courant continu.
#define directionMoteurDroit  4
#define pwmMoteurDroit  5
#define directionMoteurGauche  7
#define pwmMoteurGauche  6

// Déclarations pour le servomoteur
SoftwareServo servocam;

// Déclarations pour les capteurs de suivi de ligne
int pinCapteurLigneGauche = 0; // connexion sur entrée analogique 0
int pinCapteurLigneCentre = 1; // connexion sur entrée analogique 1
int pinCapteurLigneDroite = 2; // connexion sur entrée analogique 2

// Ce programme reçoit par liaison série des données issues d'un programme qui est exécuté sur
// l'ordinateur connecté à la carte. Si le programme ne reçoit rien pendant un temps donné,
// défini ci-dessous, un traitement particulier sera réalisé.
// Timeout de réception de données en s
#define TIMEOUT 2
unsigned long timeLastReceived = 0;
static int timedOut = 0;

// Ce programme envoie par liaison série des données à l'ordinateur connecté à la carte.
// Cadence d'envoi des données en ms
#define TSDATA 100
unsigned long tempsDernierEnvoi = 0;
unsigned long tempsCourant = 0;

// Certaines parties du programme sont exécutées à cadence fixe grâce à la bibliothèque FlexiTimer2.
// Cadence d'échantillonnage en ms
#define CADENCE_MS 10
volatile double dt = CADENCE_MS/1000.;
volatile double temps = -CADENCE_MS/1000.;

// Le gyropode est asservi grâce à plusieurs régulateurs de type PI
// On déclare ci-dessous les variables et paramètres nécessaires à l'asservissement et au régulateur
static double R = 0.045; // Rayon d'une roue
static double W = 0.14; // Largeur du gyropode
static double umax = 6.; // valeur max de la tension de commande du moteur
static double umin = -6.; // valeur min (ou max en négatif) de la tension de commande du moteur
static double vxmes = 0.; // vitesse longitudinale mesurée
static double omegames = 0.; // vitesse de rotation par rapport à la verticale mesurée
static double thetames = 0.; // angle par rapport à la verticale mesuré
static double xidotmes = 0.; // vitesse de rotation mesurée
static double Kpvx; // gain proportionnel pour l'asservissement de vitesse longitudinale
static double Kpvx1 = -29.7; // sous-gain proportionnel pour l'asservissement de vitesse longitudinale
static double Kpvx2 = 0.15; // sous-gain proportionnel pour l'asservissement de vitesse longitudinale
static double Kivx = -30.4; // gain intégral pour l'asservissement de vitesse longitudinale
static double Kpomega = -3.03; // gain proportionnel pour l'asservissement de verticalité
static double Kiomega = -36.6; // gain intégral pour l'asservissement de verticalité
static double Kpxidot = 0.; // gain proportionnel pour l'asservissement de rotation
static double Kixidot = 2.14; // gain intégral pour l'asservissement de rotation
static double commande_vx = 0.; // commande pour l'asservissement de vitesse longitudinale
static double commande_omega = 0.; // commande pour l'asservissement de verticalité
static double commande_xidot = 0.; // commande pour l'asservissement de rotation
static double P_vx = 0.; // action proportionnelle pour l'asservissement de vitesse longitudinale
static double I_vx = 0.; // action intégrale pour l'asservissement de vitesse longitudinale
static double P_omega = 0.; // action proportionnelle pour l'asservissement de verticalité
static double I_omega = 0.; // action intégrale pour l'asservissement de verticalité
static double P_xidot = 0.; // action proportionnelle pour l'asservissement de rotation
static double I_xidot = 0.; // action intégrale pour l'asservissement de rotation
static double thetaest = 0.; // angle d'inclinaison estimé par le filtre complémentaire
static double tau = 1.; // paramètre du filtre complémentaire

// Déclarations pour la commande des moteurs
static double omegaDroit = 0.;
static double omegaGauche = 0.;
static double commandeDroit = 0.;
static double commandeGauche = 0.;

// Déclarations pour les consignes de mouvement
static double vxref = 0.;
static double xidotref = 0.;
static double omegaref = 0.;
static double thetaref = 0.;

// Déclarations pour les accélérations, angles et vitesses de rotation
static int16_t axint;
static int16_t gyroint;
static double ax;
static double omega;

// Déclarations pour la tension batterie
static double tensionBatterie;

// Déclarations pour la gestion des modes "asservissement actif" et "chute"
static int started = false;
static int juststarted = true;
static int signe_ax = 0;

// Suivi de ligne
// Déclarations pour le suivi de ligne
static int ligneDetectee = 0;
static int entreeLigneCapteurGauche = 0;
static int entreeLigneCapteurCentre = 0;
static int entreeLigneCapteurDroite = 0;
static int capteurLigneGauche_int, capteurLigneGauche, capteurLigneGaucheInit;
static int capteurLigneCentre_int, capteurLigneCentre, capteurLigneCentreInit;
static int capteurLigneDroite_int, capteurLigneDroite, capteurLigneDroiteInit;
static int ligneBlanche = -1;
static double erreurPrecedente = 0.;
static double erreurSuiviLigne = 0.;

// Déclarations pour le PD de suivi de ligne
#define SEUIL 350 // seuil de détection de ligne. A régler par pas de 50
#define VXINIT 0.12 // vitesse initiale (m/s)
#define TSSUIVI 100 // période de calcul de l'asservissement de suivi de ligne (en ms)
unsigned long tempsDernierCalcul = 0;
unsigned long tempsCourantSuivi = 0;
static double commande_suivi = 0.; // commande de vitesse angulaire calculée par le PD
static double commande_suivi_avant_sat = 0.; // valeur de la commande avant la saturation (voir ci-dessous)
static double commande_suivi_max = 3.; // valeur max de la tension de commande
static double commande_suivi_min = -3.; // valeur min (ou max en négatif) de la tension de commande
static double Kp = 5.; // gain proportionnel
static double Kd = 20.; // gain dérivé
static double Kp3 = 5.; // gain proportionnel modifié
static double Kd3 = 20.; // gain dérivé modifié
static double Tf = 0.2; // constante de temps de filtrage de l'action dérivée du PD
static double P_x = 0.; // valeur de l'action proportionnelle
static double D_x = 0.; // valeur de l'action dérivée
static double D_x_temp = 0.; // pour calculs intermédiaires
static double deltaTemps = 0.1;
// Variables intermédiaires
static double ad = 0;
static double bd = 0;

// Déclaration des variables pour la réception des données sur la ligne série
#define NOMBRE_DE_CHAMPS 6 // nombre de champs de données reçues sur la liaison série
int champIndex = 0; // champ courant reçu
int valeursRecues[NOMBRE_DE_CHAMPS]; // tableau contenant les champs de données reçues
// Variables utilisées pour les données reçues
static double x1 = 0.;
static int servoref = 45;
static double Kp2 = 1.; // facteur multiplicatif sur le gain proportionnel du PD
static double Kd2 = 1.; // facteur multiplicatif sur le gain dérivé du PD
// Variables pour la vérification de cohérence des données
static int crc = 0;
static int crc_check = 0;

// Déclaration de l'objet "capteur accélérométrique et gyroscopique"
MPU6050 accelgyro;

// Initialisations
void setup(void) {
  
  // Codeur incrémental moteur droit
  pinMode(codeurDroitPinB, INPUT);      // entrée digitale pin B codeur
  digitalWrite(codeurDroitPinB, HIGH);  // activation de la résistance de pullup
  // A chaque changement de niveau de tension sur le pin A du codeur,
  // on exécute la fonction GestionInterruptionCodeurDroitPinA (définie à la fin du programme)
  attachInterrupt(codeurDroitInterruptionA, GestionInterruptionCodeurDroitPinA, CHANGE);

  // Codeur incrémental moteur gauche
  pinMode(codeurGauchePinB, INPUT);      // entrée digitale pin B codeur
  digitalWrite(codeurGauchePinB, HIGH);  // activation de la résistance de pullup
  // A chaque changement de niveau de tension sur le pin A du codeur,
  // on exécute la fonction GestionInterruptionCodeurGauchePinA (définie à la fin du programme)
  attachInterrupt(codeurGaucheInterruptionA, GestionInterruptionCodeurGauchePinA, CHANGE);
 
  // Moteur à courant continu droit
  pinMode(directionMoteurDroit, OUTPUT);
  pinMode(pwmMoteurDroit, OUTPUT);

  // Moteur à courant continu gauche
  pinMode(directionMoteurGauche, OUTPUT);
  pinMode(pwmMoteurGauche, OUTPUT);
  
  // Servomoteur
  servocam.attach(10);
  
  // Suivi de ligne
  pinMode(13, OUTPUT);

  // Liaison série.
  // ATTENTION: ne pas modifier la vitesse de transmission de la liaison série,
  Serial.begin(57600);
  Serial.flush();
  
  // Compteur d'impulsions des encodeurs
  ticksCodeurDroit = 0;
  ticksCodeurGauche = 0;
      
  // Initialisation du bus I2C
  Wire.begin(0x20);

  // Initialisation du capteur accélérométrique et gyroscopique
  //Serial.println("Initialisation de l'IMU...");
  accelgyro.initialize();

  // Test de la connexion
  //Serial.println("Test de la connexion...");
  //Serial.println(accelgyro.testConnection() ? "Connexion MPU6050 réussie" : "Connexion MPU6050 échouée");
  
  // On démarre seulement quand le gyropode dépasse la verticale
  // Pour que le gyropode démarre tout seul quand il est sur l'avant
  // et si on le redresse quand il est sur l'arrière
  signe_ax = -1;
  
  
  /* Le programme principal est constitué:
     - de la traditionnelle boucle "loop" des programmes Arduino, dans laquelle on lit 
       en permanence la liaison série pour récupérer les nouvelles consignes de mouvement
     - de la fonction "isrt", dont l'exécution est définie à cadence fixe par les deux instructions
       ci-dessous. La bonne méthode pour exécuter une fonction à cadence fixe, c'est de l'exécuter sur interruption
       (la boucle "loop" est alors interrompue régulièrement, à la cadence voulue, par un timer).
       Une mauvaise méthode serait d'utiliser la fonction Arduino "delay". En effet, la fonction "delay" définit
       un délai entre la fin des instructions qui la précèdent et le début des instructions qui la suivent, mais si
       on ne connait pas le temps d'exécution de ces instructions (ou si celui-ci est inconnu), on n'obtiendra
       jamais une exécution à cadence fixe.
       Il est nécessaire d'exécuter la fonction "isrt" à cadence fixe car cette fonction:
       - doit renvoyer des données à cadence fixe sur la liaison série, pour monitoring
       - calcule l'asservissement de verticalité et de mouvement, conçu pour fonctionner à une cadence bien spécifiée
     
  */
  FlexiTimer2::set(CADENCE_MS, 1/1000., isrt); // résolution timer = 1 ms
  FlexiTimer2::start();

}


// Boucle principale
void loop() {

  // Tant que l'asservissement de verticalité n'est pas activé
  while (started==false) {
    // Permet de ne pas entrer dans la logique de suivi de ligne
    temps = 0.;
    
    // Geeros vient de chuter
    if (juststarted==true) {
      
      // On réinitialise l'asservissement et on entre dans la boucle d'attente de redressement
      //Serial.print("\r\r\rBoucle d'attente\r\r\r");
      juststarted = false;
      
      // Réinitilisation de l'asservissement
      I_vx = 0.;
      I_omega = 0.;
      I_xidot = 0.;
      thetaest = 0.;

    }
    
    // Mesure de la pesanteur
    axint = accelgyro.getAccelerationY();
    ax = -((double)axint/16384)*9.81;

    // Si l'accélération change de signe, cela signifie que Geeros a été redressé pour le faire démarrer
    if (ax*signe_ax<0) {
      
      // Donc, on démarre
      started = true;
      //Serial.print("\r\r\rStarted\r\r\r");
      
      // Les consignes reçues sont initialisées
      x1 = VXINIT;
      
      // On réinitialise le comptage d'impulsions des encodeurs
      // On aurait pu désactiver les interruptions pendant cette phase
      // mais ce qu'on veut au final, c'est cette init
      ticksCodeurDroit = 0;
      ticksCodeurGauche = 0;
    
    }
    else {
      // On garde tout à zéro
      I_vx = 0.;
      I_omega = 0.;
      I_xidot = 0.;
      thetaest = 0.;
    }
  }


  // Réception des données sur la liaison série
  readData();

}

// Fonction excutée sur interruption
void isrt(){

  // Compteur de boucle et variable d'activation de l'asservissement
  int i;
  double en = 0.; 
  
  // Mesure de la vitesse des moteurs grâce aux codeurs incrémentaux
  
  // On calcule une moyenne glissante de la vitesse sur les Nmoy derniers échantillons
  // Nombre de ticks codeur depuis la dernière fois accumulé pour les 1à derniers échantillons
  // Ce nombre est mis à jour par les fonctions GestionInterruptionCodeurDroitPinA et GestionInterruptionCodeurDroitPinA,
  // exécutées à chaque interruption due à une impulsion sur la voie A ou B du codeur incrémental
  for (int i=0; i<Nmoy; i++) {
    ticksCodeurDroitTab[i] += ticksCodeurDroit;
  }  
  // Une fois lu, ce nombre est remis à 0 pour pouvoir l'incrémenter de nouveau sans risquer de débordement de variable
  ticksCodeurDroit = 0;
  
  // Pour l'échantillon courant, calcule de l'angle de rotation du moteur pendant la période d'échantillonnage
  codeurDroitDeltaPos = ticksCodeurDroitTab[indiceTicksCodeurDroit];
  // Remise à zéro du compteur d'impulsion codeur pour l'échantillon courant
  ticksCodeurDroitTab[indiceTicksCodeurDroit] = 0;
  
  // Mise à jour de l'indice d'échantillon courant
  indiceTicksCodeurDroit++;
  if (indiceTicksCodeurDroit==Nmoy) {
    indiceTicksCodeurDroit = 0;
  }
  
  // On calcule une moyenne glissante de la vitesse sur les Nmoy derniers échantillons
  // Nombre de ticks codeur depuis la dernière fois accumulé pour les 1à derniers échantillons
  // Ce nombre est mis à jour par les fonctions GestionInterruptionCodeurGauchePinA et GestionInterruptionCodeurGauchePinA,
  // exécutées à chaque interruption due à une impulsion sur la voie A ou B du codeur incrémental
  for (int i=0; i<Nmoy; i++) {
    ticksCodeurGaucheTab[i] += ticksCodeurGauche;
  }  
  // Une fois lu, ce nombre est remis à 0 pour pouvoir l'incrémenter de nouveau sans risquer de débordement de variable
  ticksCodeurGauche = 0;
  
  // Pour l'échantillon courant, calcule de l'angle de rotation du moteur pendant la période d'échantillonnage
  codeurGaucheDeltaPos = ticksCodeurGaucheTab[indiceTicksCodeurGauche];
  // Remise à zéro du compteur d'impulsion codeur pour l'échantillon courant
  ticksCodeurGaucheTab[indiceTicksCodeurGauche] = 0;
  
  // Mise à jour de l'indice d'échantillon courant
  indiceTicksCodeurGauche++;
  if (indiceTicksCodeurGauche==Nmoy) {
    indiceTicksCodeurGauche = 0;
  }

  // Calcul de la vitesse de rotation. C'est le nombre d'impulsions converti en radian, divisé par la période d'échantillonnage
  // On fait un calcul glissant sur Nmoy échantillons, d'où le Nmoy*dt
  // ATTENTION: le facteur 2 devant est là car avec 2 moteurs, la résolution est divisée par 2
  omegaDroit = -2*((2.*3.141592*((double)codeurDroitDeltaPos))/1632.)/(Nmoy*dt);  // en rad/s
  omegaGauche = 2*((2.*3.141592*((double)codeurGaucheDeltaPos))/1632.)/(Nmoy*dt);  // en rad/s

  
  // Vitesse de rotation mesurée par le gyroscope convertie en rad/s
  omega = -((double)gyroint/131)*3.141592/180.;

  // Acceleration longitudinale
  thetames = -((double)axint/16384);
  
  // Estimation de l'angle à partir de l'accélération horizontale (filtre complémentaire)
  thetaest = (tau*dt*omega + dt*thetames + tau*thetaest)/(dt+tau);
  
  // Test pour savoir si le gyropode est toujours debout
  if (started == true) {
    
    // Si l'angle d'inclinaison n'est pas trop prononcé
    if (abs(thetames) < 1.) {
      en = 1.;
    }
    
    // Sinon, cela signifie que Geeros est tombé
    else {
      // On désactive l'asservissement
      en = 0.;
      started = false;
      juststarted = true;
      //Serial.print("\r\r\rChute\r\r\r");
      
      // Réinitialisation pour détecter un futur redressement
      if (thetames>0) {
        signe_ax = 1;
      }
      else {
        signe_ax = -1;
      }
    }
  }
  
  // Si Geeros n'a pas encore démarré, l'asservissement est désactivé
  else {
    en = 0.;
  }

  
  // Définition des entrées de la fonction d'asservissement
  vxmes = (omegaDroit + omegaGauche)*R/2 * en;
  omegames = omega * en;
  xidotmes = (omegaDroit - omegaGauche)*R/W * en;

  /******* Calcul du PI sur vx *******/
  Kpvx = (Kpvx1 + Kpvx2/R);
  
  // Terme proportionnel (la transformation de la commande par retour d'état en PI
  // conduit à une référence nulle, d'où le 0.*vxref)
  P_vx = Kpvx * (0.*vxref - vxmes);
  
  // Calcul de la commande avant saturation
  commande_vx = P_vx + I_vx;
  
  // Terme intégral (sera utilisé lors du pas d'échantillonnage suivant)
  I_vx = I_vx + Kivx * dt * (vxref - vxmes);

  /******* Fin Calcul du PI sur vx *******/

  /******* Calcul du PI sur omega *******/  
  // Terme proportionnel
  P_omega = Kpomega * (omegaref - omegames);
  
  // Terme "intégral" (c'est en fait un terme proportionnel sur theta)
  I_omega = Kiomega * (thetaref - thetaest);
  
  // Calcul de la commande avant saturation
  commande_omega = P_omega + I_omega;
  
  /******* Fin Calcul du PI sur omega *******/
  

  /********************* Suivi de ligne **************************/
  /***************************************************************/
  
  // Le calcul de l'asservissement de suivi de ligne est sous-échantillonné
  // par rapport à l'interruption de base (par défaut, exécution une fois sur dix,
  // c'est-à-dire à 10 Hz)
  tempsCourantSuivi = millis();
  if (tempsCourantSuivi-tempsDernierCalcul > TSSUIVI) {

    // Lecture des 3 capteurs
    capteurLigneGauche_int = analogRead(pinCapteurLigneGauche);
    capteurLigneCentre_int = analogRead(pinCapteurLigneCentre);
    capteurLigneDroite_int = analogRead(pinCapteurLigneDroite);
    
    // Initialisation
    if (temps < 1.) {
      vxref = 0.;
      
      // Les variables suffixées "Init" correspondent normalement à la lecture du capteur
      // en dehors de la ligne
      capteurLigneGaucheInit = capteurLigneGauche_int;
      capteurLigneCentreInit = capteurLigneCentre_int;
      capteurLigneDroiteInit = capteurLigneDroite_int;
    }
    
    // Si la ligne n'a pas encore été détectée...
    // Si elle a été détectée une fois par tous les capteurs, on rentre plus jamais dans cette branche,
    // même si la ligne a été perdue ensuite. Cette branche sert essentiellement à déterminer la couleur
    // de la ligne (blanche sur fond noir ou noir sur fond blanc)
    // Point d'amélioration possible: faire une détermination automatique de la valeur du seuil
    if (ligneDetectee == 0) {
      vxref = VXINIT;
      
      // Si, pour la première fois, la valeur du capteur gauche (gyropode vu de dessus) est inférieure
      // à la valeur initiale (= en dehors de la ligne) moins le seuil de détection 
      if ((capteurLigneGauche_int < (capteurLigneGaucheInit - SEUIL)) && (entreeLigneCapteurGauche == 0)) {
        // Commutation d'un flag pour ne plus entrer dans ce if
        entreeLigneCapteurGauche = 1;
        // On vient de détecter une ligne blanche (les valeurs faibles du capteur correspondent à des couleurs
        // claires, les valeurs fortes correspondent à des couleurs foncées)
        ligneBlanche = 1;
      }
      // Même chose, mais détection d'une ligne noire
      if ((capteurLigneGauche_int > (capteurLigneGaucheInit + SEUIL)) && (entreeLigneCapteurGauche == 0)) {
        entreeLigneCapteurGauche = 1;
        ligneBlanche = 0;
      }
      
      // Même chose pour le capteur du centre et le capteur droit
      if ((capteurLigneCentre_int < (capteurLigneCentreInit - SEUIL)) && (entreeLigneCapteurCentre == 0)) {
        entreeLigneCapteurCentre = 1;
        erreurPrecedente = 0.01;
      }
      if ((capteurLigneCentre_int > (capteurLigneCentreInit + SEUIL)) && (entreeLigneCapteurCentre == 0)) {
        entreeLigneCapteurCentre = 1;
        erreurPrecedente = 0.01;
      }
      
      if ((capteurLigneDroite_int < (capteurLigneDroiteInit - SEUIL)) && (entreeLigneCapteurDroite == 0)) {
        entreeLigneCapteurDroite = 1;
      }
      if ((capteurLigneDroite_int > (capteurLigneDroiteInit + SEUIL)) && (entreeLigneCapteurDroite == 0)) {
        entreeLigneCapteurDroite = 1;
      }
      
      // Quand tous les capteurs ont détecté la ligne, on allume la diode de la carte Romeo
      if (entreeLigneCapteurGauche + entreeLigneCapteurCentre + entreeLigneCapteurDroite == 3) {
        ligneDetectee = 1;
        digitalWrite(13,HIGH);
      }
      
    }
    
    // On rentre dans cette branche quand la première détection a eu lieu, même si le robot perd ensuite la ligne
    else {
      
      vxref = x1;
      Kp3 = Kp * Kp2;
      Kd3 = Kd * Kd2;
      
      // Si la ligne est noire
      if (ligneBlanche == 0) {
        // Si la valeur du capteur gauche = la valeur du fond blanc + le seuil
        if (capteurLigneGauche_int > (capteurLigneGaucheInit + SEUIL)) {
          // Le capteur est sur la ligne 
          capteurLigneGauche = 1;
        }
        // Sinon, il est à côté de la ligne
        else {
          capteurLigneGauche = 0;
        }
        // Même chose avec les autres capteurs
        if (capteurLigneCentre_int > (capteurLigneCentreInit + SEUIL)) {
          capteurLigneCentre = 1;
        }
        else {
          capteurLigneCentre = 0;
        }
        if (capteurLigneDroite_int > (capteurLigneDroiteInit + SEUIL)) {
          capteurLigneDroite = 1;
        }
        else {
          capteurLigneDroite = 0;
        }
      }
      
      // Même principe que précédemment pour une ligne blanche
      else if (ligneBlanche == 1) {
        if (capteurLigneGauche_int < (capteurLigneGaucheInit - SEUIL)) {
          capteurLigneGauche = 1;
        }
        else {
          capteurLigneGauche = 0;
        }
        if (capteurLigneCentre_int < (capteurLigneCentreInit - SEUIL)) {
          capteurLigneCentre = 1;
        }
        else {
          capteurLigneCentre = 0;
        }
        if (capteurLigneDroite_int < (capteurLigneDroiteInit - SEUIL)) {
          capteurLigneDroite = 1;
        }
        else {
          capteurLigneDroite = 0;
        }
      }
                
      // Calcul de l'erreur de suivi
      erreurSuiviLigne = erreurSuivi(capteurLigneGauche, capteurLigneCentre, capteurLigneDroite);
      
      if (abs(erreurSuiviLigne) >= 0.025) {
        // Aucun capteur n'est sur la ligne, on éteint la diode
        digitalWrite(13,LOW);
      }
      else {
        digitalWrite(13,HIGH);
      }
        
      /******* Calcul du PD de suivi de ligne *******/
      // Temps entre deux mesures
      deltaTemps = dt*TSSUIVI/CADENCE_MS;
      
      // Paramètre de filtrage de l'action dérivée
      ad = Tf/(Tf+deltaTemps);
      
      // Si au moins deux capteurs sur trois sont sur la ligne, le gain dérivé
      // est divisé par deux pour éviter des à-coups de rotation trop violents
      // alors que le robot est probablement en présence d'une ligne droite
      if (abs(erreurSuiviLigne)<0.015) {
        bd = Kd3/(Tf+deltaTemps)/2;
      }
      else {
        bd = Kd3/(Tf+deltaTemps);
      }
      
      // Calcul du terme dérivé
      D_x_temp = erreurSuiviLigne - erreurPrecedente;
      D_x = ad * D_x + bd * D_x_temp;
      
            
      // Calcul du terme proportionnel
      // On ne l'active que lorsque le robot est en dehors de la ligne, pour rattraper celle-ci
      if (abs(erreurSuiviLigne)>0.025) {
        P_x = Kp3 * erreurSuiviLigne;
      }
      else {
        P_x = 0.;
      }
          
      // Calcul de la commande avant saturation
      commande_suivi_avant_sat = P_x + D_x;
      
      // Application de la saturation sur la commande
      if (commande_suivi_avant_sat > commande_suivi_max) {
        commande_suivi = commande_suivi_max;
      }
      else if (commande_suivi_avant_sat < commande_suivi_min) {
        commande_suivi = commande_suivi_min;
      }
      else {
        commande_suivi = commande_suivi_avant_sat;
      }
      
      // Détermination des valeurs qui seront les valeurs précédentes du prochain calcul
      erreurPrecedente = erreurSuiviLigne;
      tempsDernierCalcul = tempsCourantSuivi;
      /******* Fin Calcul du PD de suivi de ligne *******/
    }
  }
  
  /***************** Fin suivi de ligne **************************/
  /***************************************************************/
  
  commandeDroit = (commande_vx + commande_omega + commande_suivi) * en;
  commandeGauche = (commande_vx + commande_omega - commande_suivi) * en;

  
  // Application des commandes aux moteurs
  tensionBatterie = 7.4;
  CommandeMoteurDroit(-commandeDroit, tensionBatterie);
  CommandeMoteurGauche(-commandeGauche, tensionBatterie);
  
  // Commande servomoteur
  // servoref est défini dans la fonction readData (voir plus loin)
  servocam.write(servoref);
  SoftwareServo::refresh();
  
  // Ecriture des données sur la liaison série pour monitoring
  ecritureData();
  
  // Incrémentation du temps courant
  temps += dt;

}


void ecritureData(void) {
  // Ecriture des données en sortie tous les TSDATA millisecondes
  
  // On envoie les données s'il s'est écoulé plus de TSDATA secondes
  // entre l'instant courant et le dernier envoi
  tempsCourant = millis();
  if (tempsCourant-tempsDernierEnvoi > TSDATA) {
    Serial.print(temps);
    
    Serial.print(",");
    Serial.print(capteurLigneGauche);
    Serial.print(",");
    Serial.print(capteurLigneCentre);
    Serial.print(",");
    Serial.print(capteurLigneDroite);
    Serial.print(",");
    Serial.print(ligneDetectee);
    Serial.print(",");
    Serial.print(ligneBlanche);
    Serial.print(",");
    Serial.print(erreurSuiviLigne);
    Serial.print(",");
    Serial.print(commande_suivi);
   
    Serial.print("\r");
    Serial.print("\n");
    
    tempsDernierEnvoi = tempsCourant;
  }  
}

void readData(void) {
  // Lecture des données sur la liaison série
  // On attend une série de NOMBRE_DE_CHAMPS valeurs entières séparées par des virgules
  
  // Initialisations
  champIndex = 0;
  for (int j=0; j < NOMBRE_DE_CHAMPS; j++) {
    valeursRecues[j] = 0;
  }

  while(true) {

    // Lecture de l'accélération et de la vitesse de rotation
    // Nécessaire de faire ce qui suit ici car cette boucle infinie
    // bloque tout le reste si aucune données n'est reçue
    axint = accelgyro.getAccelerationY();
    gyroint = accelgyro.getRotationX();
  
    // Si Geeros n'a pas démarré, ce n'est pas la peine d'aller plus loin
    if (started == false) {
      break;
    }
    
    if (Serial.available()>0) { // si une donnée est reçue
      // Réinitialisation du timeout (voir plus loin)
      timeLastReceived = millis();
      timedOut = 0;
      // Lecture de la donnée
      char ch = Serial.read();
      if (ch >= '0' && ch <= '9') { // caractère ascii entre 0 et 9 ?
        // Si oui, nous stockons la valeur
        valeursRecues[champIndex] = (valeursRecues[champIndex] * 10) + (ch - '0');
      }
      else if (ch == ',') { // séparateur virgule détecté, on passe au champ suivant
        if(champIndex < NOMBRE_DE_CHAMPS-1) {
          champIndex++; // incrémentation de l'index de champ
        }
      }
      else {
        // tout caractère autre qu'un chiffre ou une virgule (le retour chariot de fin de trame par exemple)
        // stoppe l'acquisition des champs
        break; // on sort de la boucle
      }
      
    }
    // Timeout
    // Si aucune donnée n'a été reçue pendant plus de TIMEOUT ms, on conserve les valeurs
    // précédente pour que le suivi de ligne puisse continuer à se faire avec ces mêmes valeurs
    else {      
      if ((millis()-timeLastReceived > TIMEOUT*1000) && (timedOut==0)) {
        //Serial.print("\r\r\rTimeout !!!\r\r\r");
        timedOut = 1;
      }
    }    
  }
  
  // Traitement des données reçues (on extrait chaque valeur de la trame complète)
  //Serial.println(champIndex);
  if (champIndex == NOMBRE_DE_CHAMPS-1) { // si le bon nombre de champs a été reçu
      
    // Récupération de la dernière valeur (somme de contrôle, envoyée par l'émetteur)
    crc = valeursRecues[NOMBRE_DE_CHAMPS-1];
    // Calcul de la somme de contrôle (somme de toutes les autres données) côté récepteur
    crc_check = 0;
    for (int j=0; j < NOMBRE_DE_CHAMPS-1; j++) {
      crc_check += valeursRecues[j];
    }
    
    // Si les sommes de contrôle sont identiques, on récupère les données
    if (crc == crc_check) {
      // Conversion des données d'entiers en flottants
      x1 = 0.1+0.2*((double)valeursRecues[0] - 50)/100.;
      servoref = valeursRecues[2] + 45;
      Kp2 = 1.05+(3.-(double)valeursRecues[3])/20.;
      Kd2 = 1.05+(3.-(double)valeursRecues[4])/20.;
    }
    // Si les sommes de contrôle sont différentes, on fait un traitement particulier
    else {
      // Par exemple, envoi d'un message, désactivé ici
      //Serial.print("\r\r\rErreur CRC !!!");
    }
  }
  else { // on n'a pas reçu le bon nombre de champs
    // On fait un traitement particulier, pas exemple, envoi d'un message (désactivé ici)
    //Serial.print("\r\r\rErreur reception !!!\r\r\r");
  }
  
}

void CommandeMoteurDroit(double commande, double tensionBatterie) {
  // Cette fonction calcule et envoi les signaux PWM au pont en H
  // en fonction des tensions de commande et d'alimentation
  int tension_int;
  double tension;

  // L'ensemble pont en H + moteur n'est pas linéaire
  // On redresse la caractéristique avec cette fonction
  tension = redresse_commandeDroit(commande);

  // Normalisation de la tension d'alimentation par
  // rapport à la tension d'alimentation
  tension_int = (int)(255*(tension/tensionBatterie));

  // Saturation par sécurité
  if (tension_int>255) {
    tension_int = 255;
  }

  if (tension_int<-255) {
    tension_int = -255;
  }

  // Commande PWM
  if (tension_int>=0) {
    digitalWrite(directionMoteurDroit, LOW);
    analogWrite(pwmMoteurDroit, tension_int);
  }

  if (tension_int<0) {
    digitalWrite(directionMoteurDroit, HIGH);
    analogWrite(pwmMoteurDroit, -tension_int);
  }
}


double redresse_commandeDroit(double commande) {
  // Cette fonction redresse la non-linéarité de l'ensemble pont en H + moteur
  if (commande>0) {
    return(commande+0.5); 
  }
  else if (commande<0) {
    return(commande-0.5); 
  }
  else {
    return(0); 
  }
}


void CommandeMoteurGauche(double commande, double tensionBatterie) {
  // Cette fonction calcule et envoi les signaux PWM au pont en H
  // en fonction des tensions de commande et d'alimentation
  int tension_int;
  double tension;

  // L'ensemble pont en H + moteur n'est pas linéaire
  // On redresse la caractéristique avec cette fonction
  tension = redresse_commandeGauche(commande);

  // Normalisation de la tension d'alimentation par
  // rapport à la tension d'alimentation
  tension_int = (int)(255*(tension/tensionBatterie));

  // Saturation par sécurité
  if (tension_int>255) {
    tension_int = 255;
  }
  if (tension_int<-255) {
    tension_int = -255;
  }

  // Commande PWM
  if (tension_int>=0) {
    digitalWrite(directionMoteurGauche, LOW);
    analogWrite(pwmMoteurGauche, tension_int);
  }
  if (tension_int<0) {
    digitalWrite(directionMoteurGauche, HIGH);
    analogWrite(pwmMoteurGauche, -tension_int);
  }
}


double redresse_commandeGauche(double commande) {
  // Cette fonction redresse la non-linéarité de l'ensemble pont en H + moteur
  if (commande>0) {
    return(commande+0.5); 
  }
  else if (commande<0) {
    return(commande-0.5); 
  }
  else {
    return(0); 
  }
}

 
void GestionInterruptionCodeurDroitPinA() {  
  // Routine de service d'interruption attachée à la voie A du codeur incrémental droit
  // On utilise la fonction digitalReadFast2 de la librairie digitalWriteFast
  // car les lectures doivent être très rapides pour passer le moins de temps possible
  // dans cette fonction (appelée de nombreuses fois, à chaque impulsion de codeur)
  if (digitalReadFast2(codeurDroitPinA) == digitalReadFast2(codeurDroitPinB)) {
    ticksCodeurDroit++;
  }
  else {
    ticksCodeurDroit--;
  }
}


void GestionInterruptionCodeurGauchePinA() {  
  // Routine de service d'interruption attachée à la voie A du codeur incrémental gauche
  // On utilise la fonction digitalReadFast2 de la librairie digitalWriteFast
  // car les lectures doivent être très rapides pour passer le moins de temps possible
  // dans cette fonction (appelée de nombreuses fois, à chaque impulsion de codeur)
  if (digitalReadFast2(codeurGauchePinA) == digitalReadFast2(codeurGauchePinB)) {
    ticksCodeurGauche++;
  }
  else {
    ticksCodeurGauche--;
  }
}

// Calcul de l'erreur de suivi de ligne
double erreurSuivi(int capteurGauche, int capteurCentre, int capteurDroit) {
  double erreur;
  
  // Si aucun des capteurs ne voit la ligne
  if (capteurGauche + capteurCentre + capteurDroit == 0) {
    // Si l'erreur précécente était négative, il est hautement probable que le robot soit maintenant à gauche de la ligne
    if (erreurPrecedente < 0) {
      erreur = -0.03;
    }
    // Sinon, il est à droite
    else if (erreurPrecedente >= 0) {
      erreur = 0.03;
    }
  }
  
  // Si un des capteurs voit la ligne, la formule ci-dessous donne une valeur comprise entre -0.02 et 0.02
  // -0.02: seul le capteur droit (quand on regarde le robot par le dessus) est sur la ligne
  // -0.01: le capteur droit et le capteur du centre sont sur la ligne (seul le gauche est à côté)
  // 0.00: tous les capteurs sont sur la ligne
  // 0.01: le capteur gauche et le capteur du centre sont sur la ligne (seul le droit est à côté)
  // 0.02: seul le capteur gauche est sur la ligne
  else {
    erreur = (-0.02 * (double) capteurDroit + (double) (0 * capteurCentre) + 0.02 * (double) capteurGauche) / (double) (capteurGauche + capteurCentre + capteurDroit);
  }
  return(erreur);
}


